@@ -1,11 +1,15 @@
 # -*- coding=utf -*-
+# Actually, this is a furry snowflake, not a nice star
+
 from cubes.browser import *
 from cubes.common import get_logger
-from cubes.backends.sql.mapper import SnowflakeMapper, DenormalizedMapper
-from cubes.backends.sql.mapper import DEFAULT_KEY_FIELD
+from cubes.mapper import SnowflakeMapper, DenormalizedMapper
+from cubes.mapper import DEFAULT_KEY_FIELD
 import logging
 import collections
 from cubes.errors import *
+from cubes.computation import *
+from cubes.backends.sql import extensions
 
 try:
     import sqlalchemy
@@ -15,7 +19,10 @@ try:
         "sum": sql.functions.sum,
         "min": sql.functions.min,
         "max": sql.functions.max,
-        "count": sql.functions.count
+        "count": sql.functions.count,
+        "avg": extensions.avg,
+        "stddev": extensions.stddev,
+        "variance": extensions.variance
     }
 
 except ImportError:
@@ -24,9 +31,8 @@ except ImportError:
     aggregation_functions = {}
 
 __all__ = [
-    "StarBrowser",
-    "coalesce_drilldown",
-    "create_workspace"
+    "SnowflakeBrowser",
+    "QueryContext"
 ]
 
 # Required functionality checklist
@@ -58,13 +64,13 @@ __all__ = [
 #     * dimension table prefix
 #     * schema
 
-class StarBrowser(AggregationBrowser):
-    """docstring for StarBrowser"""
+class SnowflakeBrowser(AggregationBrowser):
+    """docstring for SnowflakeBrowser"""
 
-    def __init__(self, cube, connectable=None, locale=None, metadata=None, 
+    def __init__(self, cube, connectable=None, locale=None, metadata=None,
                  debug=False, **options):
-        """StarBrowser is a SQL-based AggregationBrowser implementation that 
-        can aggregate star and snowflake schemas without need of having 
+        """SnowflakeBrowser is a SQL-based AggregationBrowser implementation that
+        can aggregate star and snowflake schemas without need of having
         explicit view or physical denormalized table.
 
         Attributes:
@@ -77,12 +83,13 @@ class StarBrowser(AggregationBrowser):
         * `options` - passed to the mapper and context (see their respective
           documentation)
 
-        **Limitations:**
+        Limitations:
 
         * only one locale can be used for browsing at a time
         * locale is implemented as denormalized: one column for each language
+
         """
-        super(StarBrowser, self).__init__(cube)
+        super(SnowflakeBrowser, self).__init__(cube)
 
         if cube == None:
             raise ArgumentError("Cube for browser should not be None.")
@@ -146,9 +153,12 @@ class StarBrowser(AggregationBrowser):
 
         # TODO: add ordering (ORDER BY)
 
+        statement = self.context.denormalized_statement()
         cond = self.context.condition_for_cell(cell)
 
-        statement = self.context.denormalized_statement(whereclause=cond.condition)
+        if cond.condition is not None:
+            statement = statement.where(cond.condition)
+
         statement = paginated_statement(statement, page, page_size)
         statement = ordered_statement(statement, order, context=self.context)
 
@@ -161,7 +171,7 @@ class StarBrowser(AggregationBrowser):
 
         return ResultIterator(result, labels)
 
-    def values(self, cell, dimension, depth=None, paths=None, hierarchy=None, 
+    def values(self, cell, dimension, depth=None, paths=None, hierarchy=None,
                 page=None, page_size=None, order=None, **options):
         """Return values for `dimension` with level depth `depth`. If `depth`
         is ``None``, all levels are returned.
@@ -187,10 +197,12 @@ class StarBrowser(AggregationBrowser):
         for level in levels:
             attributes.extend(level.attributes)
 
-        cond = self.context.condition_for_cell(cell)
-        statement = self.context.denormalized_statement(whereclause=cond.condition,
-                                                        attributes=attributes,
+        statement = self.context.denormalized_statement(attributes=attributes,
                                                         include_fact_key=False)
+        cond = self.context.condition_for_cell(cell)
+
+        if cond.condition is not None:
+            statement = statement.where(cond.condition)
         statement = paginated_statement(statement, page, page_size)
         statement = ordered_statement(statement, order, context=self.context)
 
@@ -205,8 +217,9 @@ class StarBrowser(AggregationBrowser):
 
         return ResultIterator(result, labels)
 
-    def aggregate(self, cell=None, measures=None, drilldown=None, attributes=None, 
-                  page=None, page_size=None, order=None, **options):
+    def aggregate(self, cell=None, measures=None, drilldown=None,
+                  attributes=None, page=None, page_size=None, order=None,
+                  **options):
         """Return aggregated result.
 
         Number of database queries:
@@ -219,10 +232,6 @@ class StarBrowser(AggregationBrowser):
         if not cell:
             cell = Cell(self.cube)
 
-        # TODO: add ordering (ORDER BY)
-        if options.get("order_by"):
-            self.logger.warn("ordering in aggregations is not yet implemented")
-
         # TODO: add documentation
 
         # Coalesce measures - make sure that they are Attribute objects, not
@@ -233,6 +242,7 @@ class StarBrowser(AggregationBrowser):
 
         result = AggregationResult()
         result.cell = cell
+        result.measures = measures
 
         summary_statement = self.context.aggregation_statement(cell=cell,
                                                      measures=measures,
@@ -259,6 +269,14 @@ class StarBrowser(AggregationBrowser):
         ##
 
         if drilldown:
+            drilldown = levels_from_drilldown(cell, drilldown)
+
+            dim_levels = {}
+            for dim, levels in drilldown:
+                dim_levels[str(dim)] = [str(level) for level in levels]
+
+            result.levels = dim_levels
+
             statement = self.context.aggregation_statement(cell=cell,
                                                          measures=measures,
                                                          attributes=attributes,
@@ -373,6 +391,10 @@ class StarBrowser(AggregationBrowser):
 Condition = collections.namedtuple("Condition",
                                     ["attributes", "condition"])
 
+"""Aliased table information"""
+AliasedTable = collections.namedtuple("AliasedTable",
+                                    ["schema", "table", "alias"])
+
 class QueryContext(object):
 
     def __init__(self, cube, mapper, metadata, **options):
@@ -422,29 +444,64 @@ class QueryContext(object):
                     (self.schema, self.fact_name): self.fact_table
                 }
 
+        # Collect all tables and their aliases.
+        # FIXME: this should be cached per-browser/cube instance
+
+        # table_aliases contains mapping between aliased table name and real
+        # table name with alias:
+        # 
+        #       (schema, aliased_name) --> (schema, real_name, alias)
+        #
+        self.table_aliases = {
+            (self.schema, self.fact_name): (self.schema, self.fact_name, None)
+        }
+
+        # Collect all table aliases from joins detail tables
+        for join in self.mapper.joins:
+            # just ask for the table
+            table = AliasedTable(join.detail.schema,
+                                 join.detail.table,
+                                 join.alias)
+            table_alias = (join.detail.schema, join.alias or join.detail.table)
+            self.table_aliases[table_alias] = table
+
     def aggregation_statement(self, cell, measures=None,
                               attributes=None, drilldown=None):
         """Return a statement for summarized aggregation. `whereclause` is
         same as SQLAlchemy `whereclause` for
         `sqlalchemy.sql.expression.select()`. `attributes` is list of logical
         references to attributes to be selected. If it is ``None`` then all
-        attributes are used."""
+        attributes are used. `drilldown` has to be a dictionary. Use
+        `levels_from_drilldown()` to prepare correct drill-down statement."""
 
-        # TODO: do not ignore attributes
         cell_cond = self.condition_for_cell(cell)
-        attributes = cell_cond.attributes
 
-        if drilldown:
-            drilldown = coalesce_drilldown(cell, drilldown)
-            for levels in drilldown.values():
-                for level in levels:
-                    attributes |= set(level.attributes)
+        if not attributes:
+            attributes = set()
+
+            if drilldown:
+                for dim, levels in drilldown:
+                    for level in levels:
+                        attributes |= set(level.attributes)
+
+        attributes = set(attributes) | set(cell_cond.attributes)
+
+        join_expression = self.join_expression_for_attributes(attributes)
 
         # TODO: add measures as well
-        join_expression = self.join_expression_for_attributes(cell_cond.attributes)
-
         selection = []
 
+        group_by = None
+        if drilldown:
+            group_by = []
+            for dim, levels in drilldown:
+                for level in levels:
+                    columns = [self.column(attr) for attr in level.attributes
+                                                        if attr in attributes]
+                    group_by.extend(columns)
+                    selection.extend(columns)
+
+        # Measures
         if measures is None:
             measures = self.cube.measures
 
@@ -456,31 +513,22 @@ class QueryContext(object):
         # TODO: make this label configurable (should we?)
         # TODO: make presence of this configurable (shoud we?)
         rcount_label = "record_count"
-
         selection.append(sql.functions.count().label(rcount_label))
 
-        group_by = None
-
-        if drilldown:
-            group_by = []
-            for dim, levels in drilldown.items():
-                for level in levels:
-                    columns = [self.column(attr) for attr in level.attributes]
-                    group_by.extend(columns)
-                    selection.extend(columns)
-
         select = sql.expression.select(selection,
-                                    whereclause=cell_cond.condition,
                                     from_obj=join_expression,
                                     use_labels=True,
                                     group_by=group_by)
 
+        if cell_cond.condition is not None:
+            select = select.where(cell_cond.condition)
+
         return select
 
     def aggregations_for_measure(self, measure):
-        """Returns list of aggregation functions (sqlalchemy) on measure columns. 
-        The result columns are labeled as `measure` + ``_`` = `aggregation`,
-        for example: ``amount_sum`` or ``discount_min``.
+        """Returns list of aggregation functions (sqlalchemy) on measure
+        columns.  The result columns are labeled as `measure` + ``_`` =
+        `aggregation`, for example: ``amount_sum`` or ``discount_min``.
 
         `measure` has to be `Attribute` instance.
 
@@ -506,13 +554,13 @@ class QueryContext(object):
 
         return result
 
-    def denormalized_statement(self, whereclause=None, attributes=None,
-                                expand_locales=False, include_fact_key=True):
+    def denormalized_statement(self, attributes=None, expand_locales=False,
+                               include_fact_key=True):
         """Return a statement (see class description for more information) for
         denormalized view. `whereclause` is same as SQLAlchemy `whereclause`
         for `sqlalchemy.sql.expression.select()`. `attributes` is list of
-        logical references to attributes to be selected. If it is ``None`` then
-        all attributes are used.
+        logical references to attributes to be selected. If it is ``None``
+        then all attributes are used.
 
         Set `expand_locales` to ``True`` to expand all localized attributes.
         """
@@ -529,19 +577,23 @@ class QueryContext(object):
             columns.insert(0, key_column)
 
         select = sql.expression.select(columns,
-                                    whereclause=whereclause,
                                     from_obj=join_expression,
                                     use_labels=True)
 
         return select
-        
+
     def fact_statement(self, key_value):
         """Return a statement for selecting a single fact based on `key_value`"""
 
         key_column = self.fact_table.c[self.fact_key]
         condition = key_column == key_value
-        return self.denormalized_statement(whereclause=condition)
-        
+        statement = self.denormalized_statement()
+
+        if condition is not None:
+            statement = statement.where(condition)
+
+        return statement
+
 
     def join_expression_for_attributes(self, attributes, expand_locales=False):
         """Returns a join expression for `attributes`"""
@@ -567,7 +619,7 @@ class QueryContext(object):
                 raise MappingError("Detail table name should be present and should not be a fact table.")
 
             master_table = self.table(join.master.schema, join.master.table)
-            detail_table = self.table(join.detail.schema, join.detail.table, join.alias)
+            detail_table = self.table(join.detail.schema, join.alias or join.detail.table)
 
             try:
                 master_column = master_table.c[join.master.column]
@@ -608,7 +660,8 @@ class QueryContext(object):
 
             if isinstance(cut, PointCut):
                 path = cut.path
-                wrapped_cond = self.condition_for_point(dim, path)
+                wrapped_cond = self.condition_for_point(dim, path,
+                                                        cut.hierarchy)
 
                 condition = wrapped_cond.condition
                 attributes |= wrapped_cond.attributes
@@ -617,15 +670,18 @@ class QueryContext(object):
                 set_conds = []
 
                 for path in cut.paths:
-                    wrapped_cond = self.condition_for_point(dim, path)
+                    wrapped_cond = self.condition_for_point(dim, path,
+                                                            cut.hierarchy)
                     set_conds.append(wrapped_cond.condition)
                     attributes |= wrapped_cond.attributes
 
                 condition = sql.expression.or_(*set_conds)
 
             elif isinstance(cut, RangeCut):
-                # FIXME: use hierarchy
-                range_cond = self.range_condition(cut.dimension, None, cut.from_path, cut.to_path)
+                range_cond = self.range_condition(cut.dimension,
+                                                  cut.hierarchy,
+                                                  cut.from_path,
+                                                  cut.to_path)
                 condition = range_cond.condition
                 attributes |= range_cond.attributes
 
@@ -668,11 +724,11 @@ class QueryContext(object):
         return Condition(attributes,condition)
 
     def range_condition(self, dim, hierarchy, from_path, to_path):
-        """Return a condition for range (`from_path`, `to_path`). Return
-        value is a `Condition` tuple."""
-        
+        """Return a condition for a hierarchical range (`from_path`,
+        `to_path`). Return value is a `Condition` tuple."""
+
         dim = self.cube.dimension(dim)
-        
+
         lower = self.boundary_condition(dim, hierarchy, from_path, 0)
         upper = self.boundary_condition(dim, hierarchy, to_path, 1)
 
@@ -714,7 +770,7 @@ class QueryContext(object):
 
             for attr in level.attributes:
                 attributes.add(attr)
-        
+
         # Select required operator according to bound
         # 0 - lower bound
         # 1 - upper bound
@@ -732,35 +788,43 @@ class QueryContext(object):
             attributes.add(attr)
 
         condition = sql.expression.and_(*conditions)
-        attributes |= last.attributes 
+        attributes |= last.attributes
 
-        if last.condition:
-            condition = sql.expression.or_(condition, last.condition) 
+        if last.condition is not None:
+            condition = sql.expression.or_(condition, last.condition)
 
         return Condition(attributes, condition)
 
-    def table(self, schema, table_name, alias=None):
+    def table(self, schema, table_name):
         """Return a SQLAlchemy Table instance. If table was already accessed,
         then existing table is returned. Otherwise new instance is created.
 
-        If `schema` is ``None`` then browser's schema is used. If `table_name`
-        is ``None``, then fact table is used.
+        If `schema` is ``None`` then browser's default schema is used.
         """
 
-        # table_name = table_name or self.fact_name
-        aliased_name = alias or table_name
-        table_ref = (schema or self.schema, aliased_name)
-        if table_ref in self.tables:
-            return self.tables[table_ref]
+        aliased_ref = (schema or self.schema, table_name)
 
-        table = sqlalchemy.Table(table_name, self.metadata,
-                                 autoload=True, schema=schema)
+        if aliased_ref in self.tables:
+            return self.tables[aliased_ref]
 
-        self.logger.debug("registering table '%s' as '%s'" % (table_name, aliased_name))
-        if alias:
-            table = table.alias(alias)
+        # Get real table reference
+        try:
+            table_ref = self.table_aliases[aliased_ref]
+        except KeyError:
+            raise ModelError("Table with reference %s not found. "
+                             "Missing join in cube '%s'?" %
+                                    (aliased_ref, self.cube.name) )
+
+        table = sqlalchemy.Table(table_ref.table, self.metadata,
+                                 autoload=True, schema=table_ref.schema)
+
+        self.logger.debug("registering table '%s' as '%s'" % (table_ref.table,
+                                                                table_name))
+        if table_ref.alias:
+            table = table.alias(table_ref.alias)
+
+        self.tables[aliased_ref] = table
 
-        self.tables[table_ref] = table
         return table
 
     def column(self, attribute, locale=None):
@@ -796,52 +860,60 @@ class QueryContext(object):
 
         return columns
 
+class AggregatedCubeBrowser(AggregationBrowser):
+    """docstring for SnowflakeBrowser"""
 
-def coalesce_drilldown(cell, drilldown):
-    """Returns a dictionary where keys are dimensions and values are list of
-    levels to be drilled down. `drilldown` should be a list of dimensions (or
-    dimension names) or a dictionary where keys are dimension names and values
-    are level names to drill up to.
+    def __init__(self, cube, connectable=None, locale=None, metadata=None,
+                 debug=False, **options):
+        """AggregatedCubeBrowser is a SQL-based AggregationBrowser
+        implementation that uses pre-aggregated table.
 
-    For the list of dimensions or if the level is not specified, then up to
-    the next level in the cell is considered.
-    """
+        Attributes:
 
-    # TODO: consider hierarchies (currently ignored, default is used)
-    result = {}
-    depths = cell.level_depths()
+        * `cube` - browsed cube
+        * `connectable` - SQLAlchemy connectable object (engine or connection)
+        * `locale` - locale used for browsing
+        * `metadata` - SQLAlchemy MetaData object
+        * `debug` - output SQL to the logger at INFO level
+        * `options` - passed to the mapper and context (see their respective
+          documentation)
 
-    # If the drilldown is a list, convert it into a dictionary
-    if not isinstance(drilldown, dict):
-        drilldown = {dim:None for dim in drilldown}
+        """
+        super(AggregatedCubeBrowser, self).__init__(cube)
 
-    for dim, level in drilldown.items():
-        dim = cell.cube.dimension(dim)
+        if cube == None:
+            raise ArgumentError("Cube for browser should not be None.")
 
-        if level:
-            hier = dim.hierarchy()
-            index = hier.level_index(level)
-            result[dim.name] = hier[:index+1]
+        self.logger = get_logger()
+
+        self.cube = cube
+        self.locale = locale or cube.model.locale
+        self.debug = debug
+
+        if connectable is not None:
+            self.connectable = connectable
+            self.metadata = metadata or sqlalchemy.MetaData(bind=self.connectable)
+
+        # Mapper is responsible for finding corresponding physical columns to
+        # dimension attributes and fact measures. It also provides information
+        # about relevant joins to be able to retrieve certain attributes.
+
+        if options.get("use_denormalization"):
+            mapper_class = DenormalizedMapper
         else:
-            depth = depths.get(str(dim), 0)
-            result[dim.name] = drilldown_levels(dim, depth+1)
+            mapper_class = SnowflakeMapper
 
-    return result
+        self.logger.debug("using mapper %s for cube '%s' (locale: %s)" % \
+                            (str(mapper_class.__name__), cube.name, locale))
 
+        self.mapper = mapper_class(cube, locale=self.locale, **options)
+        self.logger.debug("mapper schema: %s" % self.mapper.schema)
 
-def drilldown_levels(dimension, depth, hierarchy=None):
-    """Get drilldown levels up to level at `depth`. If depth is ``None``
-    returns first level only. `dimension` has to be `Dimension` instance. """
+        # QueryContext is creating SQL statements (using SQLAlchemy). It
+        # also caches information about tables retrieved from metadata.
 
-    hier = dimension.hierarchy(hierarchy)
-    depth = depth or 0
-
-    if depth > len(hier):
-        raise ArgumentError("Hierarchy %s in dimension %s has only %d levels, "
-                         "can not drill to %d" % \
-                         (hier,dimension,len(hier),depth))
-
-    return hier[:depth]
+        self.context = QueryContext(self.cube, self.mapper,
+                                      metadata=self.metadata)
 
 def paginated_statement(statement, page, page_size):
     """Returns paginated statement if page is provided, otherwise returns
@@ -861,7 +933,9 @@ def ordered_statement(statement, order, context):
     # or as some column from joined table. Here we get the list of already
     # selected columns and derived aggregates
 
-    selection = dict(statement.columns)
+    selection = collections.OrderedDict()
+    for c in statement.columns:
+        selection[str(c)] = c
 
     # Make sure that the `order` is a list of of tuples (`attribute`,
     # `order`). If element of the `order` list is a string, then it is
@@ -907,7 +981,7 @@ def ordered_statement(statement, order, context):
             # TODO: add natural ordering for measures (may be nice)
             attribute = None
 
-        if attribute and attribute.order and name not in ordering:
+        if attribute and attribute.order and name not in order_by.keys():
             order_by[name] = order_column(column, attribute.order)
 
     return statement.order_by(*order_by.values())
@@ -949,244 +1023,3 @@ class ResultIterator(object):
 
         return dict(zip(self.labels, row))
 
-####
-# Backend related functions
-###
-def ddl_for_model(url, model, fact_prefix=None, dimension_prefix=None, schema_type=None):
-    """Create a star schema DDL for a model.
-
-    Parameters:
-
-    * `url` - database url – no connection will be created, just used by
-       SQLAlchemy to determine appropriate engine backend
-    * `cube` - cube to be described
-    * `dimension_prefix` - prefix used for dimension tables
-    * `schema_type` - ``logical``, ``physical``, ``denormalized``
-
-    As model has no data storage type information, following simple rule is
-    used:
-
-    * fact ID is an integer
-    * all keys are strings
-    * all attributes are strings
-    * all measures are floats
-
-    .. warning::
-
-        Does not respect localized models yet.
-
-    """
-    raise NotImplementedError
-
-def create_workspace(model, **options):
-    """Create workspace for `model` with configuration in dictionary
-    `options`. This method is used by the slicer server.
-
-    The options are:
-
-    Required (one of the two, `engine` takes precedence):
-
-    * `url` - database URL in form of:
-      ``backend://user:password@host:port/database``
-    * `engine` - SQLAlchemy engine - either this or URL should be provided
-
-    Optional:
-
-    * `schema` - default schema, where all tables are located (if not
-      explicitly stated otherwise)
-    * `dimension_prefix` - used by snowflake mapper to find dimension tables
-      when no explicit mapping is specified
-    * `fact_prefix` - used by the snowflake mapper to find fact table for a
-      cube, when no explicit fact table name is specified
-
-     Options for denormalized views:
-
-    * `use_denormalization` - browser will use dernormalized view instead of
-      snowflake
-    * `denormalized_view_prefix` - if denormalization is used, then this
-      prefix is added for cube name to find corresponding cube view
-    * `denormalized_view_schema` - schema wehere denormalized views are
-      located (use this if the views are in different schema than fact tables,
-      otherwise default schema is going to be used)
-    """
-    engine = options.get("engine")
-
-    if engine:
-        del options["engine"]
-    else:
-        try:
-            db_url = options["url"]
-        except KeyError:
-            raise ArgumentError("No URL or engine specified in options, "
-                                "provide at least one")
-        engine = sqlalchemy.create_engine(db_url)
-
-
-    workspace = SQLStarWorkspace(model, engine, **options)
-
-    return workspace
-
-class SQLStarWorkspace(object):
-    """Factory for browsers"""
-    def __init__(self, model, engine, **options):
-        """Create a workspace. For description of options see
-        `create_workspace()` """
-
-        super(SQLStarWorkspace, self).__init__()
-
-        self.logger = get_logger()
-
-        self.model = model
-        self.engine = engine
-        self.schema = options.get("schema")
-        self.metadata = sqlalchemy.MetaData(bind=self.engine,schema=self.schema)
-        self.options = options
-
-    def browser_for_cube(self, cube, locale=None):
-        """Creates, configures and returns a browser for a cube.
-
-        .. note::
-
-            Use `workspace.browser()` instead.
-        """
-
-        # TODO(Stiivi): make sure that we are leaking connections here
-        self.logger.info("workspace.create_browser() is depreciated, use "
-                         ".browser() instead")
-
-        return self.browser(cube, locale)
-
-    def browser(self, cube, locale=None):
-        """Returns a browser for a cube."""
-        cube = self.model.cube(cube)
-        browser = StarBrowser(cube, self.engine, locale=locale,
-                              metadata=self.metadata,
-                              **self.options)
-        return browser
-
-    def create_denormalized_view(self, cube, view_name=None, materialize=False, 
-                                 replace=False, create_index=False, 
-                                 keys_only=False, schema=None):
-        """Creates a denormalized view named `view_name` of a `cube`. If
-        `view_name` is ``None`` then view name is constructed by pre-pending
-        value of `denormalized_view_prefix` from workspace options to the cube
-        name. If no prefix is specified in the options, then view name will be
-        equal to the cube name.
-
-        Options:
-
-        * `materialize` - whether the view is materialized (a table) or
-          regular view
-        * `replace` - if `True` then existing table/view will be replaced,
-          otherwise an exception is raised when trying to create view/table
-          with already existing name
-        * `create_index` - if `True` then index is created for each key
-          attribute. Can be used only on materialized view, otherwise raises
-          an exception
-        * `keys_only` - if ``True`` then only key attributes are used in the
-          view, all other detail attributes are ignored
-        * `schema` - target schema of the denormalized view, if not specified,
-          then `denormalized_view_schema` from options is used if specified,
-          otherwise default workspace schema is used (same schema as fact
-          table schema).
-        """
-
-        cube = self.model.cube(cube)
-
-        mapper = SnowflakeMapper(cube, cube.mappings, **self.options)
-        context = QueryContext(cube, mapper, metadata=self.metadata)
-
-        key_attributes = []
-        for dim in cube.dimensions:
-            key_attributes += dim.key_attributes()
-
-        if keys_only:
-            statement = context.denormalized_statement(attributes=key_attributes, expand_locales=True)
-        else:
-            statement = context.denormalized_statement(expand_locales=True)
-
-        schema = schema or self.options.get("denormalized_view_schema") or self.schema
-
-        dview_prefix = self.options.get("denormalized_view_prefix","")
-        view_name = view_name or dview_prefix + cube.name
-
-        table = sqlalchemy.Table(view_name, self.metadata,
-                                 autoload=False, schema=schema)
-
-        preparer = self.engine.dialect.preparer(self.engine.dialect)
-        full_name = preparer.format_table(table)
-
-        if mapper.fact_name == view_name and schema == mapper.schema:
-            raise WorkspaceError("target denormalized view is the same as source fact table")
-
-        if table.exists():
-            if not replace:
-                raise WorkspaceError("View or table %s (schema: %s) already exists." % \
-                                   (view_name, schema))
-
-            inspector = sqlalchemy.engine.reflection.Inspector.from_engine(self.engine)
-            view_names = inspector.get_view_names(schema=schema)
-
-            if view_name in view_names:
-                # Table reflects a view
-                drop_statement = "DROP VIEW %s" % full_name
-                self.engine.execute(drop_statement)
-            else:
-                # Table reflects a table
-                table.drop(checkfirst=False)
-
-        if materialize:
-            create_stat = "CREATE TABLE"
-        else:
-            create_stat = "CREATE OR REPLACE VIEW"
-
-        statement = "%s %s AS %s" % (create_stat, full_name, str(statement))
-        self.logger.info("creating denormalized view %s (materialized: %s)" \
-                                            % (full_name, materialize))
-        # print("SQL statement:\n%s" % statement)
-        self.engine.execute(statement)
-
-        if create_index:
-            if not materialize:
-                raise WorkspaceError("Index can be created only on a materialized view")
-                
-            # self.metadata.reflect(schema = schema, only = [view_name] )
-            table = sqlalchemy.Table(view_name, self.metadata,
-                                     autoload=True, schema=schema)
-            self.engine.reflecttable(table)
-
-            for attribute in key_attributes:
-                label = attribute.ref()
-                self.logger.info("creating index for %s" % label)
-                column = table.c[label]
-                name = "idx_%s_%s" % (view_name, label)
-                index = sqlalchemy.schema.Index(name, column)
-                index.create(self.engine)
-
-        return statement
-
-    def validate_model(self):
-        """Validate physical representation of model. Returns a list of 
-        dictionaries with keys: ``type``, ``issue``, ``object``.
-
-        Types might be: ``join`` or ``attribute``.
-
-        The ``join`` issues are:
-
-        * ``no_table`` - there is no table for join
-        * ``duplicity`` - either table or alias is specified more than once
-
-        The ``attribute`` issues are:
-
-        * ``no_table`` - there is no table for attribute
-        * ``no_column`` - there is no column for attribute
-        * ``duplicity`` - attribute is found more than once
-
-        """
-        issues = []
-
-        for cube in self.model.cubes:
-            browser = self.browser_for_cube(cube)
-            issues += browser.validate()
-
-        return issues
