@@ -1,3 +1,5 @@
+# -*- coding=utf -*-
+
 import logging
 import json
 import decimal
@@ -25,12 +27,14 @@ __all__ = [
     "cuts_from_string",
     "string_from_cuts",
     "string_from_path",
+    "string_from_hierarchy",
     "path_from_string",
     "cut_from_string",
     "cut_from_dict",
-    "DrilldownRow",
+    "TableRow",
     "CrossTable",
-    "cross_table"
+    "cross_table",
+    "levels_from_drilldown"
 ]
 
 class AggregationBrowser(object):
@@ -53,6 +57,7 @@ class AggregationBrowser(object):
         self.cube = cube
 
     def full_cube(self):
+        # FIXME: Depreciate this method in favor of Cell(cube)
         return Cell(self.cube)
 
     def dimension_object(self, dimension):
@@ -116,14 +121,14 @@ class AggregationBrowser(object):
         """Returns a single fact from cube specified by fact key `key`"""
         raise NotImplementedError
 
-    def values(self, cell, dimension, depth=None, paths=None, 
+    def values(self, cell, dimension, depth=None, paths=None,
                hierarchy=None, **options):
         """Return values for `dimension` with level depth `depth`. If `depth`
         is ``None``, all levels are returned.
 
         .. note::
 
-            Some backends might support only default hierarchy. 
+            Some backends might support only default hierarchy.
         """
         raise NotImplementedError
 
@@ -165,6 +170,9 @@ class AggregationBrowser(object):
         Raises `cubes.ArgumentError` when there are no queries specified
         or if a query is of unknown type.
 
+        .. `formatters` is a dictionary where keys are formatter names
+        .. (arbitrary) and values are formatter instances.
+
         *Roll-up*
 
         Report queries might contain ``rollup`` specification which will
@@ -660,20 +668,12 @@ class Cell(object):
         is base if there is a point cut with path referring to the
         most detailed level of the dimension `hierarchy`."""
 
-
         hierarchy = dimension.hierarchy(hierarchy)
-
-        dim_cut = None
-        for cut in self.cuts:
-            if isinstance(cut, PointCut) and cut.dimension == dimension:
-                dim_cut = cut
-                break
-
-        depth = dim_cut.level_depth() if dim_cut else None
-
-        return depth == len(hierarchy)
-
-    # def level(self, ):
+        cut = self.point_cut_for_dimension(dimension)
+        if cut:
+            return cut.level_depth() >= len(hierarchy)
+        else:
+            return False
 
     def _filter_dimension_cuts(self, dimension, exclude=False):
         dimension = self.cube.dimension(dimension)
@@ -718,20 +718,19 @@ CUT_STRING_SEPARATOR = '|'
 DIMENSION_STRING_SEPARATOR = ':'
 PATH_STRING_SEPARATOR = ','
 RANGE_CUT_SEPARATOR = '-'
-SET_CUT_SEPARATOR = '+'
+SET_CUT_SEPARATOR = ';'
 
 """
 point: date:2004
 range: date:2004-2010
-set ?1 : date:2004+2010+2011,04+
-set ?2 : date:[2004;2010;2033,04]
+set: date:2004;2010;2011,04
 
 """
 
 def cuts_from_string(string):
     """Return list of cuts specified in `string`. You can use this function to
     parse cuts encoded in a URL.
-    
+
     Examples::
 
         date:2004
@@ -747,11 +746,11 @@ def cuts_from_string(string):
         date:2004,5-
         date:-2010
 
-    Sets are in form ``path1+path2+path3`` (none of the paths should be
+    Sets are in form ``path1;path2;path3`` (none of the paths should be
     empty)::
 
-        date:2004+2010
-        date:2004+2005,1+2010,10
+        date:2004;2010
+        date:2004;2005,1;2010,10
 
     Grammar::
 
@@ -778,33 +777,48 @@ def cuts_from_string(string):
 
 re_element = re.compile(r"^[\w,]*$")
 re_point = re.compile(r"^[\w,]*$")
-re_set = re.compile(r"^([\w,]+)(\+([\w,]+))+$")
+re_set = re.compile(r"^([\w,]+)(;([\w,]+))+$")
 re_range = re.compile(r"^([\w,]*)-([\w,]*)$")
 
 def cut_from_string(dimension, string):
     """Returns a cut from `string` with dimension `dimension. The string
     should match one of the following patterns:
-    
+
     * point cut: ``2010,2,4``
     * range cut: ``2010-2012``, ``2010,1-2012,3,5``, ``2010,1-`` (open range)
-    * set cut: ``2010+2012``, ``2010,1+2012,3,5+2012,10``
-    
-    If the `string` does not match any of the patterns, then exception is
-    raised.
+    * set cut: ``2010;2012``, ``2010,1;2012,3,5;2012,10``
+
+    If the `string` does not match any of the patterns, then ArgumentError
+    exception is raised.
+
+    `dimension` can specify a hierarchy in form ``dimension@hierarchy`` such
+    as ``date@dqmy``.
     """
+
+    pattern = r"(?P<dim>\w+)(@(?P<hier>\w+))?"
+    match = re.match(pattern, dimension)
+
+    if match:
+        d = match.groupdict()
+        dimension = d["dim"]
+        hierarchy = d["hier"]
+    else:
+        raise ArgumentError("Dimension spec '%s' does not match "
+                            "pattern 'dimension@hierarchy'" % dimension)
+
     if re_point.match(string):
-        return PointCut(dimension, path_from_string(string))
+        return PointCut(dimension, path_from_string(string), hierarchy)
     elif re_set.match(string):
         paths = map(path_from_string, string.split(SET_CUT_SEPARATOR))
-        return SetCut(dimension, paths)
+        return SetCut(dimension, paths, hierarchy)
     elif re_range.match(string):
         (from_path, to_path) = map(path_from_string, string.split(RANGE_CUT_SEPARATOR))
-        return RangeCut(dimension, from_path, to_path)
+        return RangeCut(dimension, from_path, to_path, hierarchy)
     else:
-        raise Exception("Unknown cut format (check that keys "
+        raise ArgumentError("Unknown cut format (check that keys "
                         "consist only of of alphanumeric characters and "
                         "underscore)")
-                        
+
 def cut_from_dict(desc, cube=None):
     """Returns a cut from `desc` dictionary. If `cube` is specified, then the
     dimension is looked up in the cube and set as `Dimension` instances, if
@@ -816,13 +830,14 @@ def cut_from_dict(desc, cube=None):
 
     if dim and cube:
         dim = cube.dimension(dim)
-        
+
     if cut_type == "point":
-        return PointCut(dim, desc.get("path"))
+        return PointCut(dim, desc.get("path"), desc.get("hierarchy"))
     elif cut_type == "set":
-        return SetCut(dim, desc.get("paths"))
+        return SetCut(dim, desc.get("paths"), desc.get("hierarchy"))
     elif cut_type == "range":
-        return RangeCut(dim, desc.get("from"), desc.get("to"))
+        return RangeCut(dim, desc.get("from"), desc.get("to"),
+                                desc.get("hierarchy"))
     else:
         raise ArgumentError("Unknown cut type %s" % cut_type)
 
@@ -849,11 +864,19 @@ def string_from_path(path):
     if not all(map(re_element.match, path)):
         raise ArgumentError("Can not convert path to string: "
                             "keys contain invalid characters "
-                            "(should be alpha-numeric or underscore)")
+                            "(should be alpha-numeric or underscore) '%s'"%path)
 
     string = PATH_STRING_SEPARATOR.join(path)
     return string
 
+def string_from_hierarchy(dimension, hierarchy):
+    """Returns a string in form ``dimension@hierarchy`` or ``dimension`` if
+    `hierarchy` is ``None``"""
+    if hierarchy:
+        return "%s@%s" % (str(dimension), str(hierarchy))
+    else:
+        return str(dimension)
+
 def path_from_string(string):
     """Returns a dimension point path from `string`. The path elements are
     separated by comma ``,`` character.
@@ -870,14 +893,16 @@ def path_from_string(string):
     return path
 
 class Cut(object):
-    def __init__(self, dimension):
+    def __init__(self, dimension, hierarchy=None):
         self.dimension = dimension
+        self.hierarchy = hierarchy
 
     def to_dict(self):
         """Returns dictionary representation fo the receiver. The keys are:
         `dimension`."""
         d = {
             "dimension": str(self.dimension),
+            "hierarchy": str(self.hierarchy) if self.hierarchy else None,
             "level_depth": self.level_depth()
         }
         return d
@@ -887,12 +912,15 @@ class Cut(object):
         method"""
         raise NotImplementedError
 
+    def __repr__(self):
+        return str(self.to_dict())
+
 class PointCut(Cut):
     """Object describing way of slicing a cube (cell) through point in a
     dimension"""
 
-    def __init__(self, dimension, path):
-        super(PointCut, self).__init__(dimension)
+    def __init__(self, dimension, path, hierarchy=None):
+        super(PointCut, self).__init__(dimension, hierarchy)
         self.path = path
 
     def to_dict(self):
@@ -911,24 +939,11 @@ class PointCut(Cut):
         """Return string representation of point cut, you can use it in
         URLs"""
         path_str = string_from_path(self.path)
-
-        if type(self.dimension) == str or type(self.dimension) == unicode:
-            dim_name = self.dimension
-        else:
-            dim_name = self.dimension.name
-
-        string = dim_name + DIMENSION_STRING_SEPARATOR + path_str
+        dim_str = string_from_hierarchy(self.dimension, self.hierarchy)
+        string = dim_str + DIMENSION_STRING_SEPARATOR + path_str
 
         return string
 
-    def __repr__(self):
-        if isinstance(self.dimension, basestring):
-            dim_name = self.dimension
-        else:
-            dim_name = self.dimension.name
-
-        return '{"cut": "%s", "dimension":"%s", "path": "%s"}' % ("PointCut", dim_name, self.path)
-
     def __eq__(self, other):
         if self.dimension != other.dimension:
             return False
@@ -944,8 +959,8 @@ class RangeCut(Cut):
     dimension that has ordered points. For dimensions with unordered points
     behaviour is unknown."""
 
-    def __init__(self, dimension, from_path, to_path):
-        super(RangeCut, self).__init__(dimension)
+    def __init__(self, dimension, from_path, to_path, hierarchy=None):
+        super(RangeCut, self).__init__(dimension, hierarchy)
         self.from_path = from_path
         self.to_path = to_path
 
@@ -981,25 +996,12 @@ class RangeCut(Cut):
         else:
             to_path_str = string_from_path([])
 
-        if type(self.dimension) == str or type(self.dimension) == unicode:
-            dim_name = self.dimension
-        else:
-            dim_name = self.dimension.name
-
         range_str = from_path_str + RANGE_CUT_SEPARATOR + to_path_str
-        string = dim_name + DIMENSION_STRING_SEPARATOR + range_str
+        dim_str = string_from_hierarchy(self.dimension, self.hierarchy)
+        string = dim_str + DIMENSION_STRING_SEPARATOR + range_str
 
         return string
 
-    def __repr__(self):
-        if type(self.dimension) == str:
-            dim_name = self.dimension
-        else:
-            dim_name = self.dimension.name
-
-        return '{"cut": "%s", "dimension":"%s", "from": "%s", "to": "%s"}' % \
-                    ("RangeCut", dim_name, self.from_path, self.to_path)
-
     def __eq__(self, other):
         if self.dimension != other.dimension:
             return False
@@ -1017,8 +1019,8 @@ class SetCut(Cut):
     dimension that has ordered points. For dimensions with unordered points
     behaviour is unknown."""
 
-    def __init__(self, dimension, paths):
-        super(SetCut, self).__init__(dimension)
+    def __init__(self, dimension, paths, hierarchy=None):
+        super(SetCut, self).__init__(dimension, hierarchy)
         self.paths = paths
 
     def to_dict(self):
@@ -1040,25 +1042,12 @@ class SetCut(Cut):
         for path in self.paths:
             path_strings.append(string_from_path(path))
 
-        if type(self.dimension) == str or type(self.dimension) == unicode:
-            dim_name = self.dimension
-        else:
-            dim_name = self.dimension.name
-
         set_string = SET_CUT_SEPARATOR.join(path_strings)
-        string = dim_name + DIMENSION_STRING_SEPARATOR + set_string
+        dim_str = string_from_hierarchy(self.dimension, self.hierarchy)
+        string = dim_str + DIMENSION_STRING_SEPARATOR + set_string
 
         return string
 
-    def __repr__(self):
-        if type(self.dimension) == str:
-            dim_name = self.dimension
-        else:
-            dim_name = self.dimension.name
-
-        return '{"cut": "%s", "dimension":"%s", "paths": "%s"}' % \
-                    ("SetCut", dim_name, self.paths)
-
     def __eq__(self, other):
         if self.dimension != other.dimension:
             return False
@@ -1069,27 +1058,40 @@ class SetCut(Cut):
     def __ne__(self, other):
         return not self.__eq__(other)
 
-DrilldownRow = namedtuple("DrilldownRow", ["key", "label", "path", "record"])
+TableRow = namedtuple("TableRow", ["key", "label", "path", "is_base", "record"])
 
 class AggregationResult(object):
     """Result of aggregation or drill down.
 
     Attributes:
 
+    * `cell` – cell that this result is aggregate of
     * `summary` - dictionary of summary row fields
     * `cells` - list of cells that were drilled-down
-    * `remainder` - summary of remaining cells (not yet implemented)
     * `total_cell_count` - number of total cells in drill-down (after limit,
       before pagination)
+    * `measures` – measures that were selected in aggregation
+    * `remainder` - summary of remaining cells (not yet implemented)
+    * `levels` – aggregation levels for dimensions that were used to drill-
+      down
+
+
+    .. note::
+
+        Implementors of aggregation browsers should populate `cell`,
+        `measures` and `levels` from the aggregate query.
 
     """
     def __init__(self):
         super(AggregationResult, self).__init__()
+        self.cell = None
+        self.measures = None
+        self.levels = None
+
         self.summary = {}
         self.cells = []
-        self.remainder = {}
         self.total_cell_count = None
-        self.cell = None
+        self.remainder = {}
 
     @property
     def drilldown(self):
@@ -1113,29 +1115,20 @@ class AggregationResult(object):
         d["remainder"] = self.remainder
         d["cells"] = self.cells
         d["total_cell_count"] = self.total_cell_count
+        if self.measures is not None:
+            d["measures"] = [str(m) for m in self.measures]
+        else:
+            d["measures"] = None
 
         if self.cell:
             d["cell"] = [cut.to_dict() for cut in self.cell.cuts]
         else:
             d["cell"] = None
 
+        d["levels"] = self.levels
+
         return d
 
-    def as_json(self):
-        # FIXME: Eiter depreciate this or move it into backend. Also provide
-        # option for iterable result
-
-        def default(o):
-            if type(o) == decimal.Decimal:
-                return float(o)
-            else:
-                return JSONEncoder.default(self, o)
-
-        encoder = json.JSONEncoder(default = default, indent = 4)
-        json_string = encoder.encode(self.as_dict())
-
-        return json_string
-
     def table_rows(self, dimension, depth=None):
         """Returns iterator of drilled-down rows which yields a named tuple with
         named attributes: (key, label, path, record). `depth` is last level of
@@ -1145,6 +1138,8 @@ class AggregationResult(object):
         * `key`: value of key dimension attribute at level of interest
         * `label`: value of label dimension attribute at level of interest
         * `path`: full path for the drilled-down cell
+        * `is_base`: ``True`` when dimension element is base (can not drill
+          down more)
         * `record`: all drill-down attributes of the cell
 
         Example use::
@@ -1156,10 +1151,7 @@ class AggregationResult(object):
         `TypeError` when cut for `dimension` is not `PointCut`.
         """
 
-        cut = self.cell.cut_for_dimension(dimension)
-
-        if cut and not isinstance(cut, PointCut):
-            raise TypeError("PointCut expected for drill down iterator dimension '%s' cut (was %s)" % (dimension, type(cut)))
+        cut = self.cell.point_cut_for_dimension(dimension)
 
         path = cut.path if cut else []
 
@@ -1167,30 +1159,39 @@ class AggregationResult(object):
         dimension = self.cell.cube.dimension(dimension)
         hierarchy = dimension.hierarchy()
 
+        if self.levels:
+            # Convert "levels" to a dictionary:
+            # all_levels = dict((dim, levels) for dim, levels in self.levels)
+            dim_levels = self.levels.get(str(dimension), [])
+            is_base = len(dim_levels) >= len(hierarchy)
+        else:
+            is_base = len(hierarchy) == 1
+
         if depth:
             current_level = hierarchy[depth-1]
         else:
             levels = hierarchy.levels_for_path(path, drilldown=True)
             current_level = levels[-1]
 
-        level_key = current_level.key.full_name()
+        level_key = current_level.key.ref()
         level_label = current_level.label_attribute.ref()
 
         for record in self.cells:
             drill_path = path[:] + [record[level_key]]
 
-            row = DrilldownRow(record[level_key],
+            row = TableRow(record[level_key],
                                record[level_label],
                                drill_path,
+                               is_base,
                                record)
             yield row
 
     def cross_table(self, onrows, oncolumns, measures=None):
         """
-        Creates a cross table from result's drilldown. `onrows`
-        contains list of attribute names to be placed at rows and `oncolumns`
-        contains list of attribute names to be placet at columns. `measures` is a
-        list of measures to be put into cells. If measures are not specified, then
+        Creates a cross table from result's cells. `onrows` contains list of
+        attribute names to be placed at rows and `oncolumns` contains list of
+        attribute names to be placet at columns. `measures` is a list of
+        measures to be put into cells. If measures are not specified, then
         only ``record_count`` is used.
 
         Returns a named tuble with attributes:
@@ -1209,7 +1210,27 @@ class AggregationResult(object):
 
         """
 
-        return cross_table(self.drilldown, onrows, oncolumns, measures)
+        return cross_table(self.cells, onrows, oncolumns, measures)
+
+    def __iter__(self):
+        """Return cells as iterator"""
+        return iter(self.cells)
+
+
+    def cached(self):
+        """Return shallow copy of the receiver with cached cells. If cells are
+        an iterator, they are all fetched in a list."""
+        result = AggregationResult()
+        result.cell = self.cell
+        result.measures = self.measures
+        result.levels = self.levels
+        result.summary = self.summary
+        result.total_cell_count = self.total_cell_count
+        result.remainder = self.remainder
+
+        # Cache cells from an iterator
+        result.cells = list(self.cells)
+        return result
 
 CrossTable = namedtuple("CrossTable", ["columns", "rows", "data"])
 
@@ -1260,3 +1281,103 @@ def cross_table(drilldown, onrows, oncolumns, measures=None):
         data.append(row)
 
     return CrossTable(column_hdrs, row_hdrs, data)
+
+
+def string_to_drilldown(astring):
+    """Converts `astring` into a drilldown tuple (`dimension`, `hierarchy`,
+    `level`). The string should have a format:
+    ``dimension@hierarchy:level``. Hierarchy and level are optional.
+
+    Raises `ArgumentError` when `astring` does not match expected pattern.
+    """
+
+    pattern = r"(?P<dim>\w+)(@(?P<hier>\w+))?(:(?P<level>\w+))?"
+    match = re.match(pattern, astring)
+
+    if match:
+        d = match.groupdict()
+        return (d["dim"], d["hier"], d["level"])
+    else:
+        raise ArgumentError("String '%s' does not match drilldown level "
+                            "pattern 'dim@hier:level'" % astring)
+
+def levels_from_drilldown(cell, drilldown):
+    """Converts `drilldown` into a list of levels to be used to drill down.
+    `drilldown` can be:
+
+    * list of dimensions
+    * list of dimension level specifier strings (``dimension@hierarchy:level``)
+    * list of tuples in form (`dimension`, `hierarchy`, `level`).
+
+    If `drilldown is a list of dimensions or if the level is not specified,
+    then next level in the cell is considered. The implicit next level is
+    determined from a `PointCut` for `dimension` in the `cell`.
+
+    For other types of cuts, such as range or set, "next" level is the first
+    level of hierarachy.
+
+    Returns a list of tuples: (`dimension`, `levels`) where `levels` is a list
+    of levels to be drilled down.
+
+    .. note::
+
+        For backward compatibility the function accepts a dictionary where
+        keys are dimension names and values are level names to drill up to.
+        This is argument format is depreciated.
+
+    """
+
+    result = []
+
+    # If the drilldown is a list, convert it into a dictionary
+    if isinstance(drilldown, dict):
+        logger = get_logger()
+        logger.warn("drilldown as dictionary is depreciated. Use a list of: "
+                    "(dim, hierarchy, level) instead")
+        drilldown = [(dim,None,level) for dim, level in drilldown.items()]
+
+    for obj in drilldown:
+        if isinstance(obj, basestring):
+            obj = string_to_drilldown(obj)
+        elif len(obj) != 3:
+            raise ArgumentError("Drilldown item should be either a string "
+                            " or a tuple of three elements. Is: %s" (obj, ))
+
+        dim, hier, level = obj
+        dim = cell.cube.dimension(dim)
+
+        hier = dim.hierarchy(hier)
+
+        if level:
+            index = hier.level_index(level)
+            levels = hier[:index+1]
+        else:
+            cut = cell.point_cut_for_dimension(dim)
+            if cut:
+                cut_hierarchy = dim.hierarchy(cut.hierarchy)
+                depth = cut.level_depth()
+            else:
+                cut_hierarchy = hier
+                depth = 0
+
+            if cut_hierarchy != hier:
+                raise HierarchyError("Cut hierarchy %s for dimension %s is "
+                        "different than drilldown hierarchy %s. Can not "
+                        "determine implicit next level." % \
+                                (hier, dim, cut_hierarchy))
+
+            if depth >= len(hier):
+                raise HierarchyError("Hierarchy %s in dimension %s has only "
+                                     "%d levels, can not drill to %d" % \
+                                     (hier,dim,len(hier),depth+1))
+
+            # if index + 1 >= len(hier):
+            #     raise HierarchyError("Can not drill down. Level %s is "
+            #                          "last in hierarchy %s for dimension "
+            #                          "%s" % (level, hier, dim))
+            levels = hier[:depth+1]
+
+        result.append( (dim, levels) )
+
+    return result
+
