@@ -36,14 +36,16 @@ __all__ = [
     "Hierarchy",
     "Level",
     "Attribute",
-    "simple_model"
+    "simple_model",
+    "split_aggregate_ref",
+    "aggregate_ref"
 ]
 
 DIMENSION = 1
 MEASURE = 2
 DETAIL = 3
 
-def load_model(resource, translations = None):
+def load_model(resource, translations=None):
     """Load logical model from object reference. `resource` can be an URL,
     local file path or file-like object.
 
@@ -225,7 +227,7 @@ def create_dimension(obj, dimensions=None):
             template = dimensions[obj["template"]]
         except KeyError:
             raise NoSuchDimensionError("Could not find template dimension %s" % obj["template"])
-        levels = template.levels
+        levels = copy.deepcopy(template.levels)
         hierarchies = template.hierarchies.values()
         default_hierarchy_name = template.default_hierarchy_name
         label = template.label
@@ -239,6 +241,10 @@ def create_dimension(obj, dimensions=None):
         description = None
         info = None
 
+    label = obj.get("label") or label
+    description = obj.get("description") or description
+    info = obj.get("info") or info
+
     # Levels
     # ------
 
@@ -257,7 +263,7 @@ def create_dimension(obj, dimensions=None):
         if not attributes:
             attributes = attribute_list([name])
 
-        levels = [Level(name=name, attributes=attributes)]
+        levels = [Level(name=name, label=label, attributes=attributes)]
 
 
     # Hierarchies
@@ -283,9 +289,6 @@ def create_dimension(obj, dimensions=None):
             hierarchies = [Hierarchy(**h) for h in hierarchies]
 
     default_hierarchy_name = obj.get("default_hierarchy_name") or default_hierarchy_name
-    label = obj.get("label") or label
-    description = obj.get("description") or description
-    info = obj.get("info") or info
 
     # Merge with template:
     return Dimension(name=name,
@@ -1026,7 +1029,12 @@ class Dimension(object):
 
         Dimension class is not meant to be mutable. All level attributes will
         have new dimension assigned.
+
+        Note that the dimension will claim ownership of levels and their
+        attributes. You should make sure that you pass a copy of levels if you
+        are cloning another dimension.
         """
+
         self.name = name
 
         self.label = label
@@ -1046,7 +1054,7 @@ class Dimension(object):
             hier = Hierarchy("default", self.levels)
             self.hierarchies = {"default": hier}
 
-        # Fix dimension in hierarchies
+        # Claim ownership of hierarchies
         for hier in self.hierarchies.values():
             hier.dimension = self
 
@@ -1138,7 +1146,7 @@ class Dimension(object):
             return self._default_hierarchy()
         if isinstance(obj, basestring):
             if obj not in self.hierarchies:
-                raise KeyError("No hierarchy %s in dimension %s" % (obj, self.name))
+                raise ModelError("No hierarchy %s in dimension %s" % (obj, self.name))
             return self.hierarchies[obj]
         elif isinstance(obj, Hierarchy):
             return obj
@@ -1185,13 +1193,13 @@ class Dimension(object):
 
                         return self._flat_hierarchy
                     elif len(self.levels) > 1:
-                        raise KeyError("There are no hierarchies in dimenson %s "
+                        raise ModelError("There are no hierarchies in dimenson %s "
                                        "and there are more than one level" % self.name)
                     else:
-                        raise KeyError("There are no hierarchies in dimenson %s "
+                        raise ModelError("There are no hierarchies in dimenson %s "
                                        "and there are no levels to make hierarchy from" % self.name)
                 else:
-                    raise KeyError("No default hierarchy specified in dimension '%s' " \
+                    raise ModelError("No default hierarchy specified in dimension '%s' " \
                                    "and there is more (%d) than one hierarchy defined" \
                                    % (self.name, len(self.hierarchies)))
 
@@ -1405,6 +1413,13 @@ class Hierarchy(object):
 
         return self._levels.values()
 
+    @property
+    def levels_dict(self):
+        if not self._levels:
+            self._set_levels(self._level_refs)
+
+        return self._levels
+
     def __eq__(self, other):
         if not other or type(other) != type(self):
             return False
@@ -1446,7 +1461,7 @@ class Hierarchy(object):
         extend = 1 if drilldown else 0
 
         if depth + extend > len(self.levels):
-            raise ArgumentError("Depth %d is longer than hierarchy levels %s (drilldown: %s)" % (depth, self._levels.keys(), drilldown))
+            raise HierarchyError("Depth %d is longer than hierarchy levels %s (drilldown: %s)" % (depth, self._levels.keys(), drilldown))
 
         return self.levels[0:depth+extend]
 
@@ -1458,8 +1473,8 @@ class Hierarchy(object):
         if not level:
             return self.levels[0]
 
-        index = self._levels.keys().index(str(level))
-        if index + 1 >= len(self._levels):
+        index = self.levels_dict.keys().index(str(level))
+        if index + 1 >= len(self.levels):
             return None
         else:
             return self.levels[index + 1]
@@ -1471,7 +1486,7 @@ class Hierarchy(object):
         if level is None:
             return None
 
-        index = self._levels.keys().index(str(level))
+        index = self.levels_dict.keys().index(str(level))
         if index == 0:
             return None
         else:
@@ -1481,9 +1496,9 @@ class Hierarchy(object):
         """Get order index of level. Can be used for ordering and comparing
         levels within hierarchy."""
         try:
-            return self._levels.keys().index(str(level))
+            return self.levels_dict.keys().index(str(level))
         except ValueError:
-            raise ArgumentError("Level %s is not part of hierarchy %s"
+            raise HierarchyError("Level %s is not part of hierarchy %s"
                                     % (str(level), self.name))
 
     def is_last(self, level):
@@ -1495,14 +1510,14 @@ class Hierarchy(object):
         """Rolls-up the path to the `level`. If `level` is ``None`` then path is
         rolled-up only one level.
 
-        If `level` is deeper than last level of `path` the `cubes.ArgumentError`
+        If `level` is deeper than last level of `path` the `cubes.HierarchyError`
         exception is raised. If `level` is the same as `path` level, nothing
         happens."""
 
         if level:
             last = self.level_index(level) + 1
             if last > len(path):
-                raise ArgumentError("Can not roll-up: level '%s' in dimension "
+                raise HierarchyError("Can not roll-up: level '%s' in dimension "
                                     "'%s' is deeper than deepest element "
                                     "of path %s", str(level), self.dimension.name, path)
         else:
@@ -1521,7 +1536,7 @@ class Hierarchy(object):
         path where there are no more levels to be added - no drill down
         possible."""
 
-        return path != None and len(path) == len(self._levels)
+        return path != None and len(path) == len(self.levels)
 
     def key_attributes(self):
         """Return all dimension key attributes as a single list."""
@@ -1574,22 +1589,23 @@ class Level(object):
     * `name`: level name
     * `dimension`: dimnesion the level is associated with
     * `attributes`: list of all level attributes. Raises `ModelError` when
-      `attribute list is empty.
+      `attribute` list is empty.
     * `key`: name of level key attribute (for example: ``customer_number`` for
-      customer level, ``region_code`` for region level, ``month`` for month level).
-      key will be used as a grouping field for aggregations. Key should be
-      unique within level. If not specified, then the first attribute is used as
-      key.
+      customer level, ``region_code`` for region level, ``month`` for month
+      level).  key will be used as a grouping field for aggregations. Key
+      should be unique within level. If not specified, then the first
+      attribute is used as key.
+    * `sort_key`: name of attribute that is going to be used for sorting
     * `label_attribute`: name of attribute containing label to be displayed
-      (for example: ``customer_name`` for customer level, ``region_name`` for region level,
-      ``month_name`` for month level)
+      (for example: ``customer_name`` for customer level, ``region_name`` for
+      region level, ``month_name`` for month level)
     * `label`: human readable label of the level
     * `info`: custom information dictionary, might be used to store
       application/front-end specific information
     """
 
     def __init__(self, name, attributes, dimension = None, key=None,
-                 label_attribute=None, label=None, info=None):
+                 sort_key=None, label_attribute=None, label=None, info=None):
 
         self.name = name
         self.dimension = dimension
@@ -1616,6 +1632,8 @@ class Level(object):
             else:
                 self.label_attribute = self.key
 
+        self.sort_key = self.attribute(sort_key)
+
     def __eq__(self, other):
         if not other or type(other) != type(self):
             return False
@@ -1623,8 +1641,8 @@ class Level(object):
             return False
         elif self.label_attribute != other.label_attribute:
             return False
-        # elif self.dimension != other.dimension:
-        #     return False
+        elif self.sort_key != other.sort_key:
+             return False
 
         if self.attributes != other.attributes:
             return False
@@ -1642,7 +1660,22 @@ class Level(object):
         return self.name
 
     def __repr__(self):
-        return "<level: {name: '%s', key: %s, attributes: %s}>" % (self.name, self.key, self.attributes)
+        return str(self.to_dict())
+
+    def __deepcopy__(self, memo):
+        if self.sort_key:
+            sort_key = str(self.sort_key)
+        else:
+            sort_key = None
+
+        return Level(self.name,
+                        attributes=copy.deepcopy(self.attributes,memo),
+                        key=self.key.name,
+                        sort_key=sort_key,
+                        label_attribute=self.label_attribute.name,
+                        info=copy.copy(self.info),
+                        label=copy.copy(self.label)
+                        )
 
     def to_dict(self, full_attribute_names=False, **options):
         """Convert to dictionary"""
@@ -1655,9 +1688,13 @@ class Level(object):
         if full_attribute_names:
             out.setnoempty("key", self.key.ref())
             out.setnoempty("label_attribute", self.label_attribute.ref())
+            if self.sort_key:
+                out.setnoempty("sort_key", self.sort_key.ref())
         else:
             out.setnoempty("key", self.key.name)
             out.setnoempty("label_attribute", self.label_attribute.name)
+            if self.sort_key:
+                out.setnoempty("sort_key", self.sort_key.name)
 
         array = []
         for attr in self.attributes:
@@ -1790,6 +1827,17 @@ class Attribute(object):
         else:
             self.locales = locales
 
+    def __deepcopy__(self, memo):
+        return Attribute(self.name,
+                         self.label,
+                         dimension=self.dimension,
+                         locales=copy.deepcopy(self.locales, memo),
+                         order=copy.deepcopy(self.order, memo),
+                         description=self.description,
+                         aggregations=copy.deepcopy(self.aggregations, memo),
+                         info=copy.deepcopy(self.info, memo),
+                         format=self.format)
+
     def __str__(self):
         return self.name
 
@@ -1800,8 +1848,10 @@ class Attribute(object):
         if type(other) != Attribute:
             return False
 
-        return self.name == other.name and self.label == other.label \
-                    and self.locales == other.locales
+        return self.name == other.name \
+                and self.label == other.label \
+                and self.locales == other.locales \
+                and str(self.dimension) == str(other.dimension)
 
     def __ne__(self,other):
         return not self.__eq__(other)
@@ -1902,6 +1952,35 @@ class Attribute(object):
 
         return locale
 
+def aggregate_ref(measure, aggregate):
+    """Creates a reference string for measure aggregate. Current
+    implementation joins the measure name and aggregate name with an
+    underscore character `'_'`. Use this method in in a backend to create
+    valid aggregate reference. See also `split_aggregate_ref()`"""
+
+    return "%s_%s" % (measure, aggregate)
+
+def split_aggregate_ref(measure):
+    """Splits aggregate measure reference into measure name and aggregate
+    name. Use this method in presenters to correctly get measure name and
+    aggregate name from aggregate reference that was created by
+    `aggregate_ref()` function.
+
+    If measure is `record_count` then `("record_count", None)` is returned.
+    """
+
+    measure = str(measure)
+
+    if measure == "record_count":
+        return (measure, None)
+
+    r = measure.rfind("_")
+
+    if r == -1 or r == len(measure):
+        raise ArgumentError("Invalid aggregate reference '%s'" % measure)
+
+    return (measure[:r], measure[r+1:])
+
 def localize_common(obj, trans):
     """Localize common attributes: label and description"""
     if "label" in trans:
