@@ -116,7 +116,7 @@ class ApplicationController(object):
         info = {
             "version": cubes.__version__,
             # Backward compatibility key
-            "server_version": cubes.__version__, 
+            "server_version": cubes.__version__,
             "api_version": API_VERSION
         }
         return info
@@ -167,7 +167,7 @@ class ModelController(ApplicationController):
         return self.json_response(dim.to_dict())
 
     def _cube_dict(self, cube):
-        d = cube.to_dict(expand_dimensions = True, 
+        d = cube.to_dict(expand_dimensions = True,
                          with_mappings = False,
                          full_attribute_names = True
                          )
@@ -181,6 +181,15 @@ class ModelController(ApplicationController):
         cube = self.model.cube(cube_name)
         return self.json_response(self._cube_dict(cube))
 
+    def list_cubes(self):
+        cubes = [cube.to_dict() for cube in self.model.cubes.values()]
+        return self.json_response(cubes)
+
+    def list_cube_dimensions(self, cube_name):
+        cube = self.model.cube(cube_name)
+        dimensions = [dim.to_dict() for dim in cube.dimensions]
+        return self.json_response(dimensions)
+
     def dimension_levels(self, dim_name):
         # FIXME: remove this method
         self.logger.warn("/dimension/.../levels is depreciated")
@@ -201,7 +210,7 @@ class ModelController(ApplicationController):
 
 
 class CSVGenerator(object):
-    def __init__(self, records, fields, include_header = True, 
+    def __init__(self, records, fields, include_header = True,
                 dialect=csv.excel, encoding="utf-8", **kwds):
         # Redirect output to a queue
         self.include_header = include_header
@@ -308,11 +317,13 @@ class CubesController(ApplicationController):
         self.browser = self.app.workspace.browser(self.cube, self.locale)
 
     def prepare_cell(self):
-        cut_string = self.args.get("cut")
+        cut_strings = self.args.getlist("cut")
 
-        if cut_string:
-            self.logger.debug("preparing cell from string: '%s'" % cut_string)
-            cuts = cubes.cuts_from_string(cut_string)
+        if cut_strings:
+            cuts = []
+            for cut_string in cut_strings:
+                self.logger.debug("preparing cell from string: '%s'" % cut_string)
+                cuts += cubes.cuts_from_string(cut_string)
         else:
             self.logger.debug("preparing cell as whole cube")
             cuts = []
@@ -323,20 +334,18 @@ class CubesController(ApplicationController):
         self.create_browser(cube)
         self.prepare_cell()
 
-        drilldown = self.args.getlist("drilldown")
-        dic_drilldown = {}
+        ddlist = self.args.getlist("drilldown")
 
-        # Allow dimension:level specification for drilldown
+        drilldown = []
 
-        if drilldown:
-            for drill_dim in drilldown:
-                split = drill_dim.split(":")
-                dic_drilldown[split[0]] = split[1] if len(split) >= 2 else None
+        if ddlist:
+            for ddstring in ddlist:
+                drilldown += ddstring.split("|")
 
-        result = self.browser.aggregate(self.cell, drilldown = dic_drilldown,
-                                        page = self.page,
-                                        page_size = self.page_size,
-                                        order = self.order)
+        result = self.browser.aggregate(self.cell, drilldown=drilldown,
+                                        page=self.page,
+                                        page_size=self.page_size,
+                                        order=self.order)
 
         return self.json_response(result)
 
@@ -396,7 +405,7 @@ class CubesController(ApplicationController):
             raise NotFoundError(dim_name, "dimension",
                                 message="Dimension '%s' was not found" % dim_name)
 
-        values = self.browser.values(self.cell, dimension, depth=depth, 
+        values = self.browser.values(self.cell, dimension, depth=depth,
                                      page=self.page, page_size=self.page_size)
 
         result = {
@@ -444,7 +453,7 @@ class CubesController(ApplicationController):
 
         details = self.browser.cell_details(self.cell)
         cell_dict = self.cell.to_dict()
-        
+
         for cut, detail in zip(cell_dict["cuts"], details):
             cut["details"] = detail
 
@@ -456,7 +465,7 @@ class CubesController(ApplicationController):
         self.prepare_cell()
 
         result = self.browser.cell_details(self.cell)
-        
+
         return self.json_response(result)
 
 
@@ -470,15 +479,24 @@ class SearchController(ApplicationController):
     sql_url
     search_backend: sphinx otherwise we raise exception.
 
-    """        
+    """
 
     def create_browser(self, cube_name):
-        # FIXME: remove this (?)
-        if not cube_name:
-            cube_name = self.config.get("model", "cube")
+        # FIXME: reuse? 
+        if cube_name:
+            self.cube = self.model.cube(cube_name)
+        else:
+            if self.config.has_option("model", "cube"):
+                self.logger.debug("using default cube specified in cofiguration")
+                cube_name = self.config.get("model", "cube")
+                self.cube = self.model.cube(cube_name)
+            else:
+                self.logger.debug("using first cube from model")
+                self.cube = self.model.cubes.values()[0]
+                cube_name = self.cube.name
 
-        self.cube = self.model.cube(cube_name)
-        self.browser = self.app.workspace.browser(self.cube, locale = self.locale)
+        self.browser = self.app.workspace.browser(self.cube,
+                                                  locale=self.locale)
 
         if self.config.has_option("sphinx", "host"):
             self.sphinx_host = self.config.get("sphinx","host")
@@ -492,7 +510,7 @@ class SearchController(ApplicationController):
 
     def search(self, cube):
         self.create_browser(cube)
-        
+
         if not SphinxSearcher:
             raise ServerError("Search extension cubes_search is not installed")
 
@@ -500,14 +518,14 @@ class SearchController(ApplicationController):
 
         dimension = self.args.get("dimension")
         if not dimension:
-            return RequestError("No dimension provided for search")
+            raise RequestError("No dimension provided for search")
 
         query = self.args.get("q")
         if not query:
             query = self.args.get("query")
 
         if not query:
-            return RequestError("No search query provided")
+            raise RequestError("No search query provided")
 
         zipped = self.args.get("_zip")
 
@@ -519,7 +537,7 @@ class SearchController(ApplicationController):
                     break
 
 
-        search_result = sphinx.search(query, dimension, locale_tag = locale_tag)
+        search_result = sphinx.search(query, dimension, locale_tag=locale_tag)
 
         # FIXME: remove "values" - backward compatibility key
         result = {
