@@ -2,6 +2,7 @@
 import json
 import cubes
 import logging
+import ConfigParser
 
 # Werkzeug - soft dependency
 try:
@@ -31,18 +32,31 @@ rules = Map([
                         endpoint = (controllers.ApplicationController, 'version')),
     Rule('/locales',
                         endpoint = (controllers.ApplicationController, 'locales')),
+    #
+    # Model requests
+    #
     Rule('/model',
                         endpoint = (controllers.ModelController, 'show')),
-    Rule('/model/dimension/<string:dim_name>',
-                        endpoint = (controllers.ModelController, 'dimension')),
+
+    Rule('/model/cubes',
+                        endpoint = (controllers.ModelController, 'list_cubes')),
     Rule('/model/cube',
                         endpoint = (controllers.ModelController, 'get_default_cube')),
     Rule('/model/cube/<string:cube_name>',
                         endpoint = (controllers.ModelController, 'get_cube')),
+    Rule('/model/cube/<string:cube_name>/dimensions',
+                        endpoint = (controllers.ModelController, 'list_cube_dimensions')),
+
+    Rule('/model/dimension/<string:dim_name>',
+                        endpoint = (controllers.ModelController, 'dimension')),
     Rule('/model/dimension/<string:dim_name>/levels',
                         endpoint = (controllers.ModelController, 'dimension_levels')),
     Rule('/model/dimension/<string:dim_name>/level_names',
                         endpoint = (controllers.ModelController, 'dimension_level_names')),
+
+    #
+    # Aggregation browser requests
+    #
     Rule('/cube/<string:cube>/aggregate',
                         endpoint = (controllers.CubesController, 'aggregate')),
     Rule('/cube/<string:cube>/facts',
@@ -57,9 +71,6 @@ rules = Map([
                         endpoint = (controllers.CubesController, 'cell_details')),
     Rule('/cube/<string:cube>/details',
                         endpoint = (controllers.CubesController, 'details')),
-    Rule('/cube/<string:cube>/search',
-                        endpoint = (controllers.SearchController, 'search')),
-
     # Use default cube (specified in config as: [model] cube = ... )
     Rule('/aggregate',
                         endpoint = (controllers.CubesController, 'aggregate'),
@@ -82,6 +93,12 @@ rules = Map([
     Rule('/details',
                         endpoint = (controllers.CubesController, 'details'),
                         defaults={"cube":None}),
+    #
+    # Other utility requests
+    #
+    Rule('/cube/<string:cube>/search',
+                        endpoint = (controllers.SearchController, 'search')),
+
     Rule('/search',
                         endpoint = (controllers.SearchController, 'search'),
                         defaults={"cube":None})
@@ -90,8 +107,8 @@ rules = Map([
 class Slicer(object):
 
     def __init__(self, config=None):
-        """Create a WSGI server for providing OLAP web service. You might provide ``config``
-        as ``ConfigParser`` object.
+        """Create a WSGI server for providing OLAP web service. You might
+        provide ``config`` as ``ConfigParser`` object.
         """
 
         self.config = config
@@ -142,7 +159,7 @@ class Slicer(object):
     def wsgi_app(self, environ, start_response):
         request = Request(environ)
         urls = rules.bind_to_environ(environ)
-        
+
         try:
             endpoint, params = urls.match()
 
@@ -173,13 +190,14 @@ class Slicer(object):
         return response
 
     def localized_model(self, locale=None):
-        """Tries to translate the model. Looks for language in configuration file under 
-        ``[translations]``, if no translation is provided, then model remains untouched."""
+        """Tries to translate the model. Looks for language in configuration
+        file under ``[translations]``, if no translation is provided, then
+        model remains untouched."""
 
         # FIXME: Rewrite this to make it thread safer
         if not locale:
             return self.model
-            
+
         self.logger.debug("localization to '%s' (current: '%s') requested (has: %s)" % (locale, self.model.locale, self.model_localizations.keys()))
 
         if locale in self.model_localizations:
@@ -204,11 +222,23 @@ class Slicer(object):
         else:
             raise common.RequestError("No translation for language '%s'" % locale)
 
+def create_server(config_file):
+    """Returns a WSGI server application. `config_file` is a path to an `.ini`
+    file with slicer server configuration."""
+
+    try:
+        config = ConfigParser.SafeConfigParser()
+        config.read(config_file)
+    except Exception as e:
+        raise Exception("Unable to load configuration: %s" % e)
+
+    return Slicer(config)
+
 def run_server(config):
     """Run OLAP server with configuration specified in `config`"""
     if config.has_option("server", "host"):
         host = config.get("server", "host")
-    else: 
+    else:
         host = "localhost"
 
     if config.has_option("server", "port"):
