@@ -4,10 +4,11 @@ import json
 import re
 import cubes
 import sqlalchemy
+import datetime
+
 from sqlalchemy import Table, Column, Integer, Float, String, MetaData, ForeignKey
 from sqlalchemy import create_engine
-
-from cubes.backends.sql import coalesce_physical
+from cubes.mapper import coalesce_physical
 from cubes.backends.sql.star import *
 from cubes.errors import *
 
@@ -17,8 +18,8 @@ class StarSQLTestCase(unittest.TestCase):
             "cubes": [
                 {
                     "name": "sales",
-                    "measures": [ 
-                            {"name":"amount", "aggregations":["sum", "min"]}, 
+                    "measures": [
+                            {"name":"amount", "aggregations":["sum", "min"]},
                             "discount"
                             ],
                     "dimensions" : ["date", "flag", "product"],
@@ -44,14 +45,14 @@ class StarSQLTestCase(unittest.TestCase):
                     "name": "date",
                     "levels": [
                         { "name": "year", "attributes": ["year"] },
-                        { "name": "month", "attributes": 
+                        { "name": "month", "attributes":
                                     ["month", "month_name", "month_sname"] },
                         { "name": "day", "attributes": ["id", "day"] }
                     ],
                     "hierarchy": ["year", "month", "day"]
                 },
                 { "name": "flag" },
-                { "name": "product", 
+                { "name": "product",
                     "levels": [
                         {"name": "category",
                             "attributes": ["category",
@@ -63,7 +64,7 @@ class StarSQLTestCase(unittest.TestCase):
                                             {"name": "subcategory_name", "locales": ["en", "sk"] }
                                         ]
                         },
-                        { "name": "product", 
+                        { "name": "product",
                           "attributes": [ "id",
                                           {"name": "name"}
                                         ],
@@ -72,12 +73,12 @@ class StarSQLTestCase(unittest.TestCase):
                 }
             ]
         }
-        
+
         engine = sqlalchemy.create_engine('sqlite://')
         self.connection = engine.connect()
         metadata = sqlalchemy.MetaData()
         metadata.bind = engine
-        
+
         table = Table('sales', metadata,
                         Column('id', Integer, primary_key=True),
                         Column('amount', Float),
@@ -98,7 +99,7 @@ class StarSQLTestCase(unittest.TestCase):
                         Column('month_sname', String),
                         Column('year', Integer)
                     )
-                    
+
         table = Table('dim_product', metadata,
                         Column('id', Integer, primary_key=True),
                         Column('category_id', Integer),
@@ -119,145 +120,17 @@ class StarSQLTestCase(unittest.TestCase):
 
         self.model = cubes.create_model(model_desc)
         self.cube = self.model.cube("sales")
-        self.browser = StarBrowser(self.cube,connectable=self.connection, 
+        self.browser = SnowflakeBrowser(self.cube,connectable=self.connection,
                                     dimension_prefix="dim_")
         self.browser.debug = True
         self.cube.fact = 'sales'
         self.mapper = self.browser.mapper
 
-class StarSQLAttributeMapperTestCase(StarSQLTestCase):
-    def setUp(self):
-        super(StarSQLAttributeMapperTestCase, self).setUp()
-        self.mapper.mappings = { 
-                    "product.name": "product.product_name",
-                    "product.category": "product.category_id",
-                    "subcategory.name.en": "subcategory.subcategory_name_en",
-                    "subcategory.name.sk": "subcategory.subcategory_name_sk" 
-                }
-        
-    def test_valid_model(self):
-        """Model is valid"""
-        self.assertEqual(True, self.model.is_valid())
-        
-    def test_logical_reference(self):
-
-        attr = cubes.Attribute("month",dimension=self.model.dimension("date"))
-        self.assertEqual("date.month", self.mapper.logical(attr))
-
-        attr = cubes.Attribute("category",dimension=self.model.dimension("product"))
-        self.assertEqual("product.category", self.mapper.logical(attr))
-
-        self.assertEqual(True, self.mapper.simplify_dimension_references)
-        attr = cubes.Attribute("flag",dimension=self.model.dimension("flag"))
-        self.assertEqual("flag", self.mapper.logical(attr))
-
-        attr = cubes.Attribute("measure",dimension=None)
-        self.assertEqual("measure", self.mapper.logical(attr))
-
-    def test_logical_reference_as_string(self):
-        self.assertRaises(AttributeError, self.mapper.logical, "amount")
-
-    def test_dont_simplify_dimension_references(self):
-        self.mapper.simplify_dimension_references = False
-
-        attr = cubes.Attribute("flag",dimension=self.model.dimension("flag"))
-        self.assertEqual("flag.flag", self.mapper.logical(attr))
-
-        attr = cubes.Attribute("measure",dimension=None)
-        self.assertEqual("measure", self.mapper.logical(attr))
-
-    def test_logical_split(self):
-        split = self.mapper.split_logical
-        
-        self.assertEqual(('foo', 'bar'), split('foo.bar'))
-        self.assertEqual(('foo', 'bar.baz'), split('foo.bar.baz'))
-        self.assertEqual((None, 'foo'), split('foo'))
-        
-    def assertMapping(self, expected, logical_ref, locale = None):
-        """Create string reference by concatentanig table and column name.
-        No schema is expected (is ignored)."""
-        
-        attr = self.mapper.attributes[logical_ref]
-        ref = self.mapper.physical(attr, locale)
-        sref = ref[1] + "." + ref[2]
-        self.assertEqual(expected, sref)
-
-    def test_physical_refs_dimensions(self):
-        """Testing correct default mappings of dimensions (with and without 
-        explicit default prefix) in physical references."""
-
-        # No dimension prefix
-        self.mapper.dimension_prefix = None
-        dim = self.model.dimension("product")
-        self.assertMapping("date.year", "date.year")
-        self.assertMapping("sales.flag", "flag")
-        self.assertMapping("sales.amount", "amount")
-        # self.assertEqual("fact.flag", sref("flag.flag"))
-
-        # With prefix
-        self.mapper.dimension_prefix = "dm_"
-        self.assertMapping("dm_date.year", "date.year")
-        self.assertMapping("dm_date.month_name", "date.month_name")
-        self.assertMapping("sales.flag", "flag")
-        self.assertMapping("sales.amount", "amount")
-        self.mapper.dimension_prefix = None
-
-    def test_coalesce_physical(self):
-        def assertPhysical(expected, actual, default=None):
-            ref = coalesce_physical(actual, default)
-            self.assertEqual(expected, ref)
-            
-        assertPhysical((None, "table", "column", None), "table.column")
-        assertPhysical((None, "table", "column.foo", None), "table.column.foo")
-        assertPhysical((None, "table", "column", None), ["table", "column"])
-        assertPhysical(("schema", "table", "column", None), ["schema","table", "column"])
-        assertPhysical((None, "table", "column", None), {"column":"column"}, "table")
-        assertPhysical((None, "table", "column", None), {"table":"table",
-                                                        "column":"column"})
-        assertPhysical(("schema", "table", "column", None), {"schema":"schema",
-                                                        "table":"table",
-                                                        "column":"column"})
-        assertPhysical(("schema", "table", "column", "day"), {"schema":"schema",
-                                                        "table":"table",
-                                                        "column":"column",
-                                                        "extract":"day"})
-
-    def test_physical_refs_flat_dims(self):
-        self.cube.fact = None
-        self.assertMapping("sales.flag", "flag")
-
-    def test_physical_refs_facts(self):
-        """Testing correct mappings of fact attributes in physical references"""
-
-        fact = self.cube.fact
-        self.cube.fact = None
-        self.assertMapping("sales.amount", "amount")
-        # self.assertEqual("sales.flag", sref("flag.flag"))
-        self.cube.fact = fact
-        
-    def test_physical_refs_with_mappings_and_locales(self):
-        """Testing correct mappings of mapped attributes and localized 
-        attributes in physical references"""
-
-        # Test defaults
-        self.assertMapping("dim_date.month_name", "date.month_name")
-        self.assertMapping("dim_category.category_name_en", "product.category_name")
-        self.assertMapping("dim_category.category_name_sk", "product.category_name", "sk")
-        self.assertMapping("dim_category.category_name_en", "product.category_name", "de")
-
-        # Test with mapping
-        self.assertMapping("dim_product.product_name", "product.name")
-        self.assertMapping("dim_product.category_id", "product.category")
-        self.assertMapping("dim_product.product_name", "product.name", "sk")
-        self.assertMapping("dim_category.subcategory_name_en", "product.subcategory_name")
-        self.assertMapping("dim_category.subcategory_name_sk", "product.subcategory_name", "sk")
-        self.assertMapping("dim_category.subcategory_name_en", "product.subcategory_name", "de")
-
 
 class QueryContextTestCase(StarSQLTestCase):
     def setUp(self):
         super(QueryContextTestCase, self).setUp()
-        
+
     def test_denormalize(self):
         statement = self.browser.context.denormalized_statement()
         cols = [column.name for column in statement.columns]
@@ -268,73 +141,76 @@ class QueryContextTestCase(StarSQLTestCase):
         statement = self.browser.context.denormalized_statement(expand_locales=True)
         cols = [column.name for column in statement.columns]
         self.assertEqual(20, len(cols))
-        
-    def test_coalesce_drilldown(self):
+
+    def test_levels_from_drilldown(self):
         cell = cubes.Cell(self.cube)
         dim = self.cube.dimension("date")
+        l_year = dim.level("year")
+        l_month = dim.level("month")
+        l_day = dim.level("day")
 
-        drilldown = {"date": "year"}
-        expected = {"date": [dim.level("year")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        drilldown = [("date", None, "year")]
+        expected = [(dim, [l_year])]
+        self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
         drilldown = ["date"]
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
-        drilldown = {"date": "day"}
-        expected = {"date": [dim.level("year"), dim.level("month"), dim.level("day")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        drilldown = [("date", None, "year")]
+        expected = [(dim, [l_year])]
+        self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
         # Try "next level"
 
         cut = cubes.PointCut("date", [2010])
         cell = cubes.Cell(self.cube, [cut])
 
-        drilldown = {"date": "year"}
-        expected = {"date": [dim.level("year")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        drilldown = [("date", None, "year")]
+        expected = [(dim, [l_year])]
+        self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
         drilldown = ["date"]
-        expected = {"date": [dim.level("year"), dim.level("month")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        expected = [(dim, [l_year, l_month])]
+        self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
         # Try with range cell
 
-        cut = cubes.RangeCut("date", [2009], [2010])
-        cell = cubes.Cell(self.cube, [cut])
+        # cut = cubes.RangeCut("date", [2009], [2010])
+        # cell = cubes.Cell(self.cube, [cut])
 
-        drilldown = ["date"]
-        expected = {"date": [dim.level("year"), dim.level("month")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        # drilldown = ["date"]
+        # expected = [(dim, [l_year, l_month])]
+        # self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
-        drilldown = {"date":"year"}
-        expected = {"date": [dim.level("year")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        # drilldown = [("date", None, "year")]
+        # expected = [(dim, [l_year])]
+        # self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
-        cut = cubes.RangeCut("date", [2009], [2010, 1])
-        cell = cubes.Cell(self.cube, [cut])
+        # cut = cubes.RangeCut("date", [2009], [2010, 1])
+        # cell = cubes.Cell(self.cube, [cut])
 
-        drilldown = ["date"]
-        expected = {"date": [dim.level("year"), dim.level("month"), dim.level("day")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        # drilldown = ["date"]
+        # expected = [(dim, [l_year, l_month, l_day])]
+        # self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
         # Try "last level"
 
         cut = cubes.PointCut("date", [2010, 1,2])
         cell = cubes.Cell(self.cube, [cut])
 
-        drilldown = {"date": "day"}
-        expected = {"date": [dim.level("year"), dim.level("month"), dim.level("day")]}
-        self.assertEqual(expected, coalesce_drilldown(cell, drilldown))
+        drilldown = [("date", None, "day")]
+        expected = [(dim, [l_year, l_month, l_day])]
+        self.assertEqual(expected, cubes.levels_from_drilldown(cell, drilldown))
 
         drilldown = ["date"]
-        expected = {"date": [dim.level("year"), dim.level("month")]}
-        self.assertRaises(ArgumentError, coalesce_drilldown, cell, drilldown)
-        
+        expected = [(dim, [l_year, l_month])]
+        self.assertRaises(HierarchyError, cubes.levels_from_drilldown, cell, drilldown)
+
 
 class JoinsTestCase(StarSQLTestCase):
     def setUp(self):
         super(JoinsTestCase, self).setUp()
-        
+
         self.joins = [
                 {"master":"fact.date_id", "detail": "date.id"},
                 {"master":["fact", "product_id"], "detail": "product.id"},
@@ -343,7 +219,7 @@ class JoinsTestCase(StarSQLTestCase):
                 {"master":"subcategory.category_id", "detail": "category.id"}
             ]
         self.mapper._collect_joins(self.joins)
-        
+
     def test_basic_joins(self):
         relevant = self.mapper.relevant_joins([[None,"date"]])
         self.assertEqual(1, len(relevant))
@@ -360,17 +236,17 @@ class JoinsTestCase(StarSQLTestCase):
         self.assertEqual(1, len(relevant))
         self.assertEqual("date", relevant[0].detail.table)
         self.assertEqual("contract_date", relevant[0].alias)
-        
+
     def test_snowflake(self):
         relevant = self.mapper.relevant_joins([[None,"subcategory"]])
-        
+
         self.assertEqual(2, len(relevant))
         test = sorted([r.detail.table for r in relevant])
         self.assertEqual(["product","subcategory"], test)
         self.assertEqual([None, None], [r.alias for r in relevant])
 
         relevant = self.mapper.relevant_joins([[None,"category"]])
-        
+
         self.assertEqual(3, len(relevant))
         test = sorted([r.detail.table for r in relevant])
         self.assertEqual(["category", "product","subcategory"], test)
@@ -382,7 +258,7 @@ class StarValidationTestCase(StarSQLTestCase):
     def test_validate(self):
         result = self.browser.validate_model()
         self.assertEqual(0, len(result))
-        
+
 class StarSQLBrowserTestCase(StarSQLTestCase):
     def setUp(self):
         super(StarSQLBrowserTestCase, self).setUp()
@@ -448,7 +324,7 @@ class StarSQLBrowserTestCase(StarSQLTestCase):
     def table(self, name):
         return sqlalchemy.Table(name, self.metadata,
                                 autoload=True)
-    
+
     def test_get_fact(self):
         """Get single fact"""
         self.assertEqual(True, self.mapper.simplify_dimension_references)
@@ -469,7 +345,7 @@ class StarSQLBrowserTestCase(StarSQLTestCase):
         values = list(self.browser.values(None,"product",3))
         self.assertIsNotNone(values)
         self.assertEqual(10, len(values))
-        
+
     def test_cut_details(self):
         cut = cubes.PointCut("date", [2012])
         details = self.browser.cut_details(cut)
@@ -496,7 +372,7 @@ class StarSQLBrowserTestCase(StarSQLTestCase):
         cell = cubes.Cell( self.cube, [cubes.PointCut("product", [10])] )
         details = self.browser.cell_details(cell)
         self.assertEqual(1, len(details))
-        self.assertEqual([[{"product.category":10, 
+        self.assertEqual([[{"product.category":10,
                            "product.category_name":"Things",
                            "_key":10,
                            "_label": "Things"}]], details)
@@ -511,7 +387,7 @@ class StarSQLBrowserTestCase(StarSQLTestCase):
                 [{"product.category":10, "product.category_name": "Things",
                   "_key":10, "_label": "Things"}]
                         ], details)
-        
+
     def test_aggregation_for_measures(self):
         context = self.browser.context
 
@@ -536,7 +412,7 @@ class StarSQLBrowserTestCase(StarSQLTestCase):
         keys = sorted(result.summary.keys())
         self.assertEqual(2, len(keys))
         self.assertEqual(["discount_sum", "record_count"], keys)
-        
+
     @unittest.skip("not implemented")
     def test_aggregate_measure_only(self):
         """Aggregation result should: SELECT from fact only"""
@@ -571,14 +447,195 @@ class StarSQLBrowserTestCase(StarSQLTestCase):
         pass
 
 
-    # def test_measure_selection(self):
-        
+class HierarchyTestCase(unittest.TestCase):
+    def setUp(self):
+        model = {
+            "cubes": [
+                {
+                    "name":"cube",
+                    "dimensions": ["date"],
+                    "joins": [
+                        {"master":"date_id", "detail":"dim_date.id"}
+                    ]
+                }
+            ],
+            "dimensions": [
+                {
+                    "name": "date",
+                    "levels": [
+                        {"name":"year"},
+                        {"name":"quarter"},
+                        {"name":"month"},
+                        {"name":"week"},
+                        {"name":"day"}
+                    ],
+                    "hierarchies": [
+                        {"name": "ymd", "levels":["year", "month", "day"]},
+                        {"name": "ym", "levels":["year", "month"]},
+                        {"name": "yqmd", "levels":["year", "quarter", "month", "day"]},
+                        {"name": "ywd", "levels":["year", "week", "day"]}
+                    ],
+                    "default_hierarchy_name": "ymd"
+                }
+            ]
+        }
+
+        self.model = cubes.create_model(model)
+
+        engine = create_engine("sqlite:///")
+        metadata = MetaData(bind=engine)
+        d_table = Table("dim_date", metadata,
+                        Column('id', Integer, primary_key=True),
+                        Column('year', Integer),
+                        Column('quarter', Integer),
+                        Column('month', Integer),
+                        Column('week', Integer),
+                        Column('day', Integer))
+
+        f_table = Table("ft_cube", metadata,
+                        Column('id', Integer, primary_key=True),
+                        Column('date_id', Integer))
+        metadata.create_all()
+
+        start_date = datetime.date(2000, 1, 1)
+        end_date = datetime.date(2001, 1,1)
+        delta = datetime.timedelta(1)
+        date = start_date
+
+        d_insert = d_table.insert()
+        f_insert = f_table.insert()
+
+        i = 1
+        while date < end_date:
+            record = {
+                        "id": int(date.strftime('%Y%m%d')),
+                        "year": date.year,
+                        "quarter": (date.month-1)//3+1,
+                        "month": date.month,
+                        "week": int(date.strftime("%U")),
+                        "day": date.day
+                    }
+
+            engine.execute(d_insert.values(record))
+
+            # For each date insert one fact record
+            record = {"id": i,
+                      "date_id": record["id"]
+                      }
+            engine.execute(f_insert.values(record))
+            date = date + delta
+            i += 1
+
+        self.cube = self.model.cube("cube")
+        self.browser = SnowflakeBrowser(self.cube,
+                                        connectable=engine,
+                                        dimension_prefix="dim_",
+                                        fact_prefix="ft_")
+        self.browser.debug = True
+
+    def test_cell(self):
+        cell = cubes.Cell(self.cube)
+        result = self.browser.aggregate(cell)
+        self.assertEqual(366, result.summary["record_count"])
+
+        cut = cubes.PointCut("date", [2000, 2])
+        cell = cubes.Cell(self.cube, [cut])
+        result = self.browser.aggregate(cell)
+        self.assertEqual(29, result.summary["record_count"])
+
+        cut = cubes.PointCut("date", [2000, 2], hierarchy="ywd")
+        cell = cubes.Cell(self.cube, [cut])
+        result = self.browser.aggregate(cell)
+        self.assertEqual(7, result.summary["record_count"])
+
+        cut = cubes.PointCut("date", [2000, 1], hierarchy="yqmd")
+        cell = cubes.Cell(self.cube, [cut])
+        result = self.browser.aggregate(cell)
+        self.assertEqual(91, result.summary["record_count"])
+
+    def test_drilldown(self):
+        cell = cubes.Cell(self.cube)
+        result = self.browser.aggregate(cell, drilldown=["date"])
+        self.assertEqual(1, result.total_cell_count)
+
+        result = self.browser.aggregate(cell, drilldown=["date:month"])
+        self.assertEqual(12, result.total_cell_count)
+
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", None, "month")])
+        self.assertEqual(12, result.total_cell_count)
+
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", None, "day")])
+        self.assertEqual(366, result.total_cell_count)
+
+        # Test year-quarter-month-day
+        hier = self.model.dimension("date").hierarchy("yqmd")
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", "yqmd", "day")])
+        self.assertEqual(366, result.total_cell_count)
+
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", "yqmd", "quarter")])
+        self.assertEqual(4, result.total_cell_count)
+
+    def test_range_drilldown(self):
+        cut = cubes.RangeCut("date", [2000, 1], [2000,3])
+        cell = cubes.Cell(self.cube, [cut])
+        result = self.browser.aggregate(cell, drilldown=["date"])
+        # This should test that it does not drilldown on range
+        self.assertEqual(1, result.total_cell_count)
+
+    def test_implicit_level(self):
+        cut = cubes.PointCut("date", [2000])
+        cell = cubes.Cell(self.cube, [cut])
+
+        result = self.browser.aggregate(cell, drilldown=["date"])
+        self.assertEqual(12, result.total_cell_count)
+        result = self.browser.aggregate(cell, drilldown=["date:month"])
+        self.assertEqual(12, result.total_cell_count)
+
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", None, "month")])
+        self.assertEqual(12, result.total_cell_count)
+
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", None, "day")])
+        self.assertEqual(366, result.total_cell_count)
+
+    def test_hierarchy_compatibility(self):
+        cut = cubes.PointCut("date", [2000])
+        cell = cubes.Cell(self.cube, [cut])
+
+        self.assertRaises(HierarchyError, self.browser.aggregate,
+                            cell, drilldown=[("date", "yqmd", None)])
+
+        cut = cubes.PointCut("date", [2000], hierarchy="yqmd")
+        cell = cubes.Cell(self.cube, [cut])
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", "yqmd", None)])
+
+        self.assertEqual(4, result.total_cell_count)
+
+        cut = cubes.PointCut("date", [2000], hierarchy="yqmd")
+        cell = cubes.Cell(self.cube, [cut])
+        self.assertRaises(HierarchyError, self.browser.aggregate,
+                            cell, drilldown=[("date", "ywd", None)])
+
+        cut = cubes.PointCut("date", [2000], hierarchy="ywd")
+        cell = cubes.Cell(self.cube, [cut])
+        result = self.browser.aggregate(cell,
+                                        drilldown=[("date", "ywd", None)])
+
+        self.assertEqual(54, result.total_cell_count)
+
+
 def suite():
     suite = unittest.TestSuite()
-    suite.addTest(unittest.makeSuite(StarSQLAttributeMapperTestCase))
     suite.addTest(unittest.makeSuite(JoinsTestCase))
     suite.addTest(unittest.makeSuite(StarSQLBrowserTestCase))
     suite.addTest(unittest.makeSuite(StarValidationTestCase))
     suite.addTest(unittest.makeSuite(QueryContextTestCase))
+    suite.addTest(unittest.makeSuite(HierarchyTestCase))
 
     return suite
